import enum


def compute_pdu_checksum(pdu: list):
    # todo: move to twseia.packets module
    if not isinstance(pdu, list):
        raise ValueError(f'pdu is not list, {pdu}')
    _checksum = 0
    for x in list(pdu):
        _checksum ^= x
    return _checksum


def convert_enum_to_value_list(enum_type):
    if isinstance(enum_type, enum.EnumMeta):
        return [e.value for e in list(enum_type)]
    else:
        raise Exception(f'Object class type invalid, {type(enum_type)}')


__all__ = ['compute_pdu_checksum', 'convert_enum_to_value_list']



# -*- coding: utf-8 -*-
import enum


class SATypeIDEnum(enum.IntEnum):
    """SA類別碼, TAISEIA Spec. Table_10"""
    REGISTER = 0x00
    """註冊用"""
    AIR_CONDITIONER = 0x01
    """冷氣機"""
    REFRIGERATOR = 0x02
    """電冰箱"""
    WATCHING_MACHINE = 0x03
    """洗衣機"""
    DEHUMIDIFIER = 0x04
    """除濕機"""
    TELEVISION = 0x05
    """電視機"""
    DRYING_MACHINE = 0x06
    """烘衣機"""
    HEAD_PUMP_WATER_HEATER = 0x07
    """熱泵熱水器"""
    AIR_CLEANER = 0x08
    """空氣清淨機"""
    ELECTRONIC_POT = 0x09
    """電子鍋"""
    OPEN_DRINK_MACHINE = 0x0A
    """開飲機"""
    INDUCTION_COOKER = 0x0B
    """電磁爐"""
    DISH_WASHER = 0x0C
    """烘碗機"""
    MICROWAVE_OVEN = 0x0D
    """微波爐"""
    FULL_HEAT_SWITCH = 0x0E
    """全熱交換器"""
    FAN = 0x0F
    """電扇"""
    GAS_WATER_HEATER = 0x10
    """燃氣熱水器"""
    LAMP = 0x11
    """燈具"""
    # FAN_SAMPO = 0x13
    SMART_METER_GATEWAY = 0xE0
    """智慧電表閘道器"""
    GENERAL_DEVICE = 0xF0
    """通用裝置"""


class SARegisterServiceIDEnum(enum.IntEnum):
    """ SA註冊用服務類別, TAISEIA Spec. Table_14"""
    REGISTRATION = 0X00
    """註冊服務"""
    READ_CLASS_ID = 0X01
    """讀取裝置類別代碼(Class ID) """
    READ_PROTOCOL_VERSION = 0X02
    """讀取TaiSEIA 101裝置監控通訊協定版本值 """
    RESERVED = 0X03
    """保留, 保留供未來擴充功能使用"""
    READ_TYPE_ID = 0X04
    """讀取SA類別代碼(Type ID) """
    READ_BRAND = 0X05
    """讀取SA廠牌(Brand) """
    READ_MODEL = 0X06
    """讀取SA型號(Model) """
    READ_SUPPORTED_SERVICES = 0X07
    """讀取SA可提供之所有支援服務規格 """
    READ_CURRENT_SERVICES_STATES = 0X08
    """讀取SA目前所有服務狀態值"""


class SAServiceIOMode(enum.IntEnum):
    """SA服務類別讀/寫(R/W)功能規格"""
    READ = 0x00
    READ_WRITE = 0x01


class SAClassID(enum.IntEnum):
    """SA裝置類別: 0：家庭電器、1：發電設備、2：儲能設備、3：感測設備。"""
    HOME_DEVICE = 0x00
    POWER_GENERATION_DEVICE = 0x01
    ENERGY_STORAGE_DEVICE = 0x02
    SENSOR_DEVICE = 0x03


class SAPacketDataLenType(enum.IntEnum):
    """多位元組資料型態SA：0代表一般資料型態SA、1代表多位元組資料型態SA"""
    FIXED_LEN = 0
    DYNAMIC_LEN = 1



__title__ = 'twseia'
__description__ = """A Python library for Taiwan Smart Energy Industry Association (TaiSEIA) protocol."""
__url__ = 'https://twseia.readthedocs.io/'
__version__ = '0.0.1'
# __build__ = 0x022501
__author__ = 'Lee Shiueh'
__author_email__ = 'lee.shiueh@gmail.com'
__license__ = 'Apache 2.0'
# __copyright__ = 'Copyright 2021'



import logging
from .constants import SAServiceIOMode, SAClassID
from .constants import SATypeIDEnum
from .constants import SARegisterServiceIDEnum
from .constants import SAPacketDataLenType
from .utils import compute_pdu_checksum
from .services import SADataValueType
from .devices import AirConditioner
from .devices import Dehumidifier

logger = logging.getLogger(__name__)


class SARequestPacket:
    def to_pdu(self):
        raise NotImplementedError


class SAResponsePacket:
    @classmethod
    def from_pdu(cls, pdu: list):
        raise NotImplementedError


class _BasePacket:
    len = None
    type_id = None
    io_mode_id = None
    service_id = None
    data_type_id = None
    data_bytes = None
    checksum = None

    def to_json(self):
        payload = {
            'len': self.len,
            'type_id': self.type_id,
            'io_mode_id': self.io_mode_id,
            'service_id': self.service_id,
            'data_bytes': self.data_bytes
        }
        if self.data_type_id is not None:
            payload['data_type_id'] = self.data_type_id
        return payload

    @classmethod
    def raise_if_packet_len_checksum_invalid(cls, pdu: list):
        if not isinstance(pdu, list):
            raise ValueError(f'pdu invalid, {pdu}')

        if pdu[0] != len(pdu):
            raise ValueError(f'pdu len check invalid, {pdu}')

        if pdu[-1] != compute_pdu_checksum(pdu=pdu[:-1]):
            raise ValueError(f'pdu checksum invalid, {pdu}')

    @classmethod
    def from_general_pdu(cls, pdu: list):
        cls.raise_if_packet_len_checksum_invalid(pdu=pdu)
        _pdu = list(pdu)
        packet = cls()
        packet.len = _pdu[0]
        packet.type_id = _pdu[1]
        packet.io_mode_id = _pdu[1] >> 7
        packet.service_id = _pdu[2] & 0x7F
        packet.data_bytes = _pdu[3:-1]
        packet.checksum = _pdu[-1]
        return packet

    @classmethod
    def from_data_type_pdu(cls, pdu: list):
        cls.raise_if_packet_len_checksum_invalid(pdu=pdu)
        _pdu = list(pdu)
        packet = cls()
        packet.len = _pdu[0]
        packet.type_id = _pdu[1]
        packet.io_mode_id = _pdu[2] >> 7
        packet.service_id = _pdu[2] & 0x7F
        packet.data_type_id = _pdu[3]
        packet.data_bytes = _pdu[4:-1]
        packet.checksum = _pdu[-1]
        return packet

    def to_general_pdu(self):
        pdu = [self.type_id, (self.io_mode_id << 7) | (self.service_id & 0x7F)] + self.data_bytes
        _len = len(pdu) + 2
        pdu.insert(0, _len)
        checksum = compute_pdu_checksum(pdu)
        pdu.append(checksum)
        return pdu

    def to_data_type_pdu(self):
        pdu = [self.type_id, (self.io_mode_id << 7) | (self.service_id & 0x7F), self.data_type_id] + self.data_bytes
        _len = len(pdu) + 2
        pdu.insert(0, len(pdu))
        checksum = compute_pdu_checksum(pdu)
        pdu.append(checksum)
        return pdu


class SAInfoRequestPacket(_BasePacket, SARequestPacket):
    """TAISEIA Spec. Graph_34"""

    @classmethod
    def from_pdu(cls, pdu: list):
        return super(SAInfoRequestPacket, cls).from_general_pdu(pdu=pdu)

    def to_pdu(self):
        return super(SAInfoRequestPacket, self).to_general_pdu()

    @classmethod
    def create(cls, sa_info_type: SARegisterServiceIDEnum = SARegisterServiceIDEnum.REGISTRATION):
        packet = cls()
        packet.len = 6
        packet.type_id = 0x00
        packet.io_mode_id = SAServiceIOMode.READ.value
        packet.service_id = sa_info_type.value
        packet.data_bytes = [0xFF, 0xFF]
        return packet


class SAInfoDevStatesPacket(_BasePacket, SAResponsePacket):
    @classmethod
    def from_pdu(cls, pdu: list):
        pass


class SAInfoRegisterPacket(_BasePacket, SAResponsePacket):
    class_id = None
    major_ver = None  # protocol major version
    minor_ver = None  # protocol minor version
    fragment_offset = None  # fragment offset
    brand = None
    model = None
    services = []

    @classmethod
    def from_pdu(cls, pdu: list):
        cls.raise_if_packet_len_checksum_invalid(pdu=pdu)
        _pdu = list(pdu)
        packet = cls()
        packet.len = _pdu[0]
        assert _pdu[1] == 0x00
        packet.data_type_id = _pdu[2] >> 7  # PacketDataUnitType
        packet.class_id = _pdu[2] & 0x0F
        packet.major_ver = _pdu[3]
        packet.minor_ver = _pdu[4]
        packet.fragment_offset = _pdu[5]
        packet.type_id = SATypeIDEnum(int.from_bytes(_pdu[6:8], 'big'))
        n_zero = _pdu[8:].index(0)
        packet.brand = bytearray(_pdu[8:8 + n_zero]).decode("utf-8")
        n_start = 8 + n_zero + 1
        n_zero = _pdu[n_start:].index(0)
        packet.model = bytearray(_pdu[n_start:n_start + n_zero]).decode("utf-8")
        n_start = n_start + n_zero + 1
        n = n_start
        packet.services = []
        is_fixed_len_pdu = True if packet.data_type_id == SAPacketDataLenType.FIXED_LEN else False
        while n < len(_pdu)-1:
            if is_fixed_len_pdu:
                _len = 3
            else:
                data_type_id = _pdu[n + 1]
                _len = SADataValueType.query_data_type_bytes_len(data_type_id=data_type_id)

            # logger.debug(f'parsing dev service pdu {_pdu[n:n + _len]}')
            if packet.type_id == SATypeIDEnum.AIR_CONDITIONER:
                service = AirConditioner.convert_dev_specific_service(
                    pdu=_pdu[n:n + _len],
                    is_fixed_len_pdu=is_fixed_len_pdu)
            elif packet.type_id == SATypeIDEnum.DEHUMIDIFIER:
                service = Dehumidifier.convert_dev_specific_service(
                    pdu=_pdu[n:n + _len],
                    is_fixed_len_pdu=is_fixed_len_pdu)
            else:
                raise NotImplementedError

            packet.services.append(service)
            n += _len
        return packet

    def to_json(self):
        return {
            'class_id': self.class_id,
            'class_name': SAClassID(self.class_id).name,
            'type_id': self.type_id,
            'type_name': SATypeIDEnum(self.type_id).name,
            'version': (self.major_ver, self.minor_ver),
            'fragment_offset': self.fragment_offset,
            'brand': self.brand,
            'model': self.model,
            'services': [service.to_json() for service in self.services]
        }


class SAInfoResponsePacket(_BasePacket):
    """TAISEIA Spec. Graph_35"""

    @classmethod
    def from_pdu(cls, pdu: list):
        packet = super(SAInfoResponsePacket, cls).from_general_pdu(pdu=pdu)
        if packet.service_id == SARegisterServiceIDEnum.REGISTRATION:
            if packet.io_mode_id != SAServiceIOMode.READ:
                raise ValueError(f'Packet io_mode_id {packet.io_mode_id} invalid')
            return SAInfoRegisterPacket.from_pdu(pdu=pdu)
        elif packet.service_id == SARegisterServiceIDEnum.READ_CURRENT_SERVICES_STATES:
            return SAInfoDevStatesPacket.from_pdu(pdu=pdu)
        else:
            raise NotImplementedError


class SAStateReadRequestPacket(_BasePacket):
    """TAISEIA Spec. Graph_36"""

    @classmethod
    def create(cls, type_id: int, service_id: int):
        packet = cls()
        packet.len = 0x06
        packet.type_id = type_id
        packet.io_mode_id = SAServiceIOMode.READ.value
        packet.service_id = service_id
        packet.data_bytes = [0xFF, 0xFF]
        return packet

    def to_pdu(self):
        return self.to_general_pdu()


class SAStateReadResponsePacket(_BasePacket):
    """TAISEIA Spec. Graph_37"""

    @classmethod
    def from_pdu(cls, pdu: list):
        return cls.from_general_pdu(pdu=pdu)


class SAStateReadMultiRequestPacket:
    """TAISEIA Spec. Graph_38"""
    len = None
    type_id = None  # Table_10
    service_id = None  # Appendix, SA specific
    data_type_id = None
    data_bytes = None  # [0xFF] * N
    checksum = None


class SAStateReadMultiResponsePacket:
    """TAISEIA Spec. Graph_39"""
    len = None
    type_id = None  # Table_10
    service_id = None  # Appendix, SA specific
    data_type_id = None
    data_bytes = None
    checksum = None


class SASensorReadMultiRequestPacket:
    """TAISEIA Spec. Graph_40"""
    len = None
    type_id = None  # Table_10
    service_id = None  # Appendix, SA specific
    data_type_id = None
    data_bytes = None  # [0xFF] * N
    checksum = None


class SASensorReadMultiResponsePacket:
    """TAISEIA Spec. Graph_41"""
    len = None
    type_id = None  # Table_10
    service_id = None  # Appendix, SA specific
    data_type_id = None
    fragment_offset = None
    data_bytes = None
    checksum = None


class SAStateWriteRequestPacket(_BasePacket):
    """TAISEIA Spec. Graph_43"""

    @classmethod
    def create(cls, type_id: int, service_id: int, value: int):
        packet = cls()
        packet.len = 6
        packet.type_id = type_id
        packet.service_id = 0x80 & service_id
        packet.data_bytes = value.to_bytes(2, 'big')
        return packet

    def to_pdu(self):
        return self.to_general_pdu()


class SAStateWriteResponsePacket(SAStateWriteRequestPacket):

    @classmethod
    def from_pdu(cls, pdu: list):
        return cls.from_general_pdu(pdu=pdu)


class SAStateWriteMultiRequestPacket:
    """TAISEIA Spec. Graph_44"""
    len = None
    type_id = None  # Table_10
    service_id = None  # Appendix, SA specific
    data_type_id = None
    data_bytes = None  # [0xFF] * N
    checksum = None


__all__ = [
    'SAInfoRequestPacket',
    'SAInfoResponsePacket',
    'SAStateReadRequestPacket',
    'SAInfoRegisterPacket',
    'SAInfoDevStatesPacket',
    'SAStateReadResponsePacket',
    'SAStateWriteRequestPacket',
    'SAStateWriteResponsePacket',
    'SASensorReadMultiResponsePacket',
]



import enum


class SADataValueType(enum.IntEnum):
    ENUM16 = 0x01
    ENUM16_BIT = 0x06

    UNIT8 = 0x0a
    UINT16 = 0x0b
    UINT32 = 0x0c
    UINT64 = 0x0d

    INT8 = 0x0f
    INT16 = 0x10
    INT32 = 0x11

    TIME_MD = 0x14
    TIME_HM = 0x15
    TIME_MS = 0x16
    TIME_YMDHMS = 0x17
    TIME_YMDHM = 0x18

    STR = 0x20


class SAServiceBase:
    io_mode_id = None
    service_id = None
    data_type_id = None
    data_bytes = None
    name = None

    @classmethod
    def from_fixed_len_pdu(cls, pdu: list):
        if not isinstance(pdu, list):
            raise ValueError(f'pdu type invalid, {pdu}')
        if len(pdu) != 3:
            raise ValueError(f'pdu len invalid, {pdu}')
        _pdu = list(pdu)
        service = cls()
        service.io_mode_id = _pdu[0] >> 7
        service.service_id = _pdu[0] & 0x7F
        service.data_bytes = _pdu[1:]
        return service

    @classmethod
    def from_dynamic_len_pdu(cls, pdu: list):
        if not isinstance(pdu, list):
            raise ValueError(f'pdu type invalid, {pdu}')
        if len(pdu) >= 3:
            raise ValueError(f'pdu len invalid, {pdu}')
        _pdu = list(pdu)
        service = cls()
        service.io_mode_id = _pdu[0] >> 7
        service.service_id = _pdu[0] & 0x7F
        service.data_type_id = _pdu[1]
        service.data_bytes = _pdu[2:]
        return service

    def to_json(self):
        return {
            'io_mode_id': self.io_mode_id,
            'service_id': self.service_id,
            'data_bytes': self.data_bytes,
            'name': self.name
        }


class SAService:
    @classmethod
    def from_service(cls, service: SAServiceBase) -> SAServiceBase:
        raise NotImplementedError


class Enum16Service(SAServiceBase):
    ID = 0x01

    def read_value(self):
        return int.from_bytes(self.data_bytes[-2:], 'big')


class Enum16BitService(SAServiceBase):
    ID = 0x06

    def read_bit(self, bit_index: int) -> int:
        assert (bit_index <= 15) and (bit_index >= 0)
        if 0 <= bit_index <= 7:
            mask = 1 << bit_index
            return (self.data_bytes[0] & mask) >> bit_index
        elif 8 <= bit_index <= 15:
            mask = 1 << (bit_index - 8)
            return (self.data_bytes[1] & mask) >> (bit_index - 8)

    def read_value(self):
        return '0b{:016b}'.format(int.from_bytes(self.data_bytes, 'big'))


class UInt8Service(SAServiceBase):
    ID = 0x0A

    def read_value(self):
        return int.from_bytes(self.data_bytes[-1:], 'big')

    def read_max(self):
        return int.from_bytes(self.data_bytes[-2:-1], 'big')

    def read_min(self):
        return int.from_bytes(self.data_bytes[-1:], 'big')


class UInt16Service(SAServiceBase):
    ID = 0x0B

    def read_value(self):
        return int.from_bytes(self.data_bytes[-2:], 'big')

    def read_max(self):
        return int.from_bytes(self.data_bytes[-4:-2], 'big')

    def read_min(self):
        return int.from_bytes(self.data_bytes[-2:], 'big')


class UInt32Service(SAServiceBase):
    ID = 0x0C

    def read_value(self):
        return int.from_bytes(self.data_bytes[-4:], 'big')

    def read_max(self):
        return int.from_bytes(self.data_bytes[-8:-4], 'big')

    def read_min(self):
        return int.from_bytes(self.data_bytes[-4:], 'big')


class UInt64Service(SAServiceBase):
    ID = 0x0D

    def read_value(self):
        return int.from_bytes(self.data_bytes[-8:], 'big')

    def read_max(self):
        return int.from_bytes(self.data_bytes[-16:-8], 'big')

    def read_min(self):
        return int.from_bytes(self.data_bytes[-8:], 'big')


class Int8Service(SAServiceBase):
    ID = 0x0F

    def read_value(self):
        return int.from_bytes(self.data_bytes[-1:], 'big', signed=True)

    def read_max(self):
        return int.from_bytes(self.data_bytes[-2:-1], 'big', signed=True)

    def read_min(self):
        return int.from_bytes(self.data_bytes[-1:], 'big', signed=True)


class Int16Service(SAServiceBase):
    ID = 0x10

    def read_value(self):
        return int.from_bytes(self.data_bytes[-2:], 'big', signed=True)

    def read_max(self):
        return int.from_bytes(self.data_bytes[-4:-2], 'big', signed=True)

    def read_min(self):
        return int.from_bytes(self.data_bytes[-2:], 'big', signed=True)


class Int32Service(SAServiceBase):
    ID = 0x11

    def read_value(self):
        return int.from_bytes(self.data_bytes[-4:], 'big', signed=True)

    def read_max(self):
        return int.from_bytes(self.data_bytes[-4:-2], 'big', signed=True)

    def read_min(self):
        return int.from_bytes(self.data_bytes[-2:], 'big', signed=True)


class MDService(SAServiceBase):
    ID = 0x14

    def read_value(self):
        return (int.from_bytes(self.data_bytes[-2:1], 'big'),
                int.from_bytes(self.data_bytes[-1:], 'big'))

    def read_month(self):
        return self.read_value()[0]

    def read_day(self):
        return self.read_value()[1]


class HMService(SAServiceBase):
    ID = 0x15

    def read_value(self):
        return (int.from_bytes(self.data_bytes[-2:1], 'big'),
                int.from_bytes(self.data_bytes[-1:], 'big'))

    def read_hour(self):
        return self.read_value()[0]

    def read_minute(self):
        return self.read_value()[1]


class MSService(SAServiceBase):
    ID = 0x16

    def read_value(self):
        return (int.from_bytes(self.data_bytes[-2:1], 'big'),
                int.from_bytes(self.data_bytes[-1:], 'big'))

    def read_minute(self):
        return self.read_value()[0]

    def read_second(self):
        return self.read_value()[1]


class YMDHMSService(SAServiceBase):
    ID = 0x17

    def read_value(self):
        raise NotImplementedError


class YMDHMService(SAServiceBase):
    ID = 0x18

    def read_value(self):
        raise NotImplementedError


class StringService(SAServiceBase):
    ID = 0x20

    def read_value(self):
        raise NotImplementedError


def query_sa_data_type_dynamic_len(data_type_id: int):
    if data_type_id in [SADataValueType.ENUM16,
                        SADataValueType.ENUM16_BIT,
                        SADataValueType.UNIT8,
                        SADataValueType.INT8,
                        SADataValueType.TIME_MD,
                        SADataValueType.TIME_HM,
                        SADataValueType.TIME_MS]:
        return 2
    elif data_type_id in [SADataValueType.UINT16,
                          SADataValueType.INT16]:
        return 4
    elif data_type_id in [SADataValueType.TIME_YMDHM]:
        return 5
    elif data_type_id in [SADataValueType.TIME_YMDHMS]:
        return 6
    elif data_type_id in [SADataValueType.UINT32,
                          SADataValueType.INT32]:
        return 8
    elif data_type_id in [SADataValueType.UINT64]:
        return 16
    elif data_type_id in [SADataValueType.STR]:
        return -1
    else:
        raise ValueError(f'data_type_id invalid, {data_type_id}')


__all__ = [
    'query_sa_data_type_dynamic_len',
    'SADataValueType',
    'SAServiceBase',
    'Enum16Service',
    'Enum16BitService',
    'UInt8Service',
    'UInt16Service',
    'UInt32Service',
    'UInt64Service',
    'Int8Service',
    'Int16Service',
    'Int32Service',
    'MDService',
    'HMService',
    'MSService'
]



"""SmartApplication: A Python driver for TaiSEIA protocol device control via serial port (via USB, RS485 or RS232)."""

from tests.sample_pdus import kHITACHI_AC_RAD_50NK_REGISTER_PDU
from .constants import SARegisterServiceIDEnum
from .constants import SATypeIDEnum
from .packets import SAInfoRequestPacket
from .packets import SAInfoRegisterPacket
from .packets import SAStateReadRequestPacket
from .packets import SAStateReadResponsePacket
from .packets import SAStateWriteRequestPacket
from .packets import SAStateWriteResponsePacket


class SmartApplication:
    """SmartApplication(SA) class for talking to TaiSEIA Smart Application device.

    Uses TaiSEIA protocol via RS232

    Attributes:
        port (str): The serial port name, for example ``/dev/ttyUSB0`` (Linux),
        ``/dev/tty.usbserial`` (OS X) or ``COM4`` (Windows).
    """
    _device = None

    def __init__(self, port='/dev/ttyUSB0'):
        """"""
        self.port = port
        # self.register()

    @property
    def services(self) -> list:
        """SA Service command `SAService` supported list."""
        if isinstance(self._device, SAInfoRegisterPacket):
            return self._device.services
        else:
            return []

    def request(self, payload: list) -> list:
        """Send a TaiSEIA protocol request bytes packet for SA.

        todo: Send bytes via real RS232.

        """
        req = SAInfoRequestPacket.from_pdu(payload)
        if req.type_id == 0x00 and req.service_id == SARegisterServiceIDEnum.REGISTRATION:
            # return SAInfoResponsePacket.from_pdu(pdu=kHITACHI_AC_RAD_50NK_REGISTER_PDU)
            return kHITACHI_AC_RAD_50NK_REGISTER_PDU
        else:
            if req.type_id == SATypeIDEnum.AIR_CONDITIONER:
                if payload == [0x06, 0x01, 0x00, 0xff, 0xff, 0x07]:
                    return [0x06, 0x01, 0x00, 0x00, 0x00, 0x07]
                elif payload == [0x06, 0x01, 0x01, 0xff, 0xff, 0x06]:
                    return [0x06, 0x01, 0x01, 0x00, 0x04, 0x02]
                elif payload == [0x06, 0x01, 0x02, 0xff, 0xff, 0x05]:
                    return [0x06, 0x01, 0x02, 0x00, 0x01, 0x04]
                elif payload == [0x06, 0x01, 0x03, 0xff, 0xff, 0x04]:
                    return [0x06, 0x01, 0x03, 0x00, 0x19, 0x1d]
                elif payload == [0x06, 0x01, 0x04, 0xff, 0xff, 0x03]:
                    return [0x06, 0x01, 0x04, 0x00, 0x19, 0x1a]
                elif payload == [0x06, 0x01, 0x06, 0xff, 0xff, 0x01]:
                    return [0x06, 0x01, 0x06, 0x00, 0x00, 0x01]
                elif payload == [0x06, 0x01, 0x0b, 0xff, 0xff, 0x0c]:
                    return [0x06, 0x01, 0x0b, 0x00, 0x00, 0x0c]
                elif payload == [0x06, 0x01, 0x0c, 0xff, 0xff, 0x0b]:
                    return [0x06, 0x01, 0x0c, 0x00, 0x00, 0x0b]
                elif payload == [0x06, 0x01, 0x17, 0xff, 0xff, 0x10]:
                    return [0x06, 0x01, 0x17, 0x00, 0x00, 0x10]
                elif payload == [0x06, 0x01, 0x1a, 0xff, 0xff, 0x1d]:
                    return [0x06, 0x01, 0x1a, 0x00, 0x00, 0x1d]
                elif payload == [0x06, 0x01, 0x1b, 0xff, 0xff, 0x1c]:
                    return [0x06, 0x01, 0x1b, 0x00, 0x00, 0x1c]
                elif payload == [0x06, 0x01, 0x1d, 0xff, 0xff, 0x1a]:
                    return [0x06, 0x01, 0x1d, 0x00, 0x00, 0x1a]
                elif payload == [0x06, 0x01, 0x1e, 0xff, 0xff, 0x19]:
                    return [0x06, 0x01, 0x1e, 0x00, 0x00, 0x19]
                elif payload == [0x06, 0x01, 0x21, 0xff, 0xff, 0x26]:
                    return [0x06, 0x01, 0x21, 0x00, 0x11, 0x37]
                elif payload == [0x06, 0x01, 0x28, 0xff, 0xff, 0x2f]:
                    return [0x06, 0x01, 0x28, 0x0f, 0x80, 0xa0]
                elif payload == [0x06, 0x01, 0x29, 0xff, 0xff, 0x2e]:
                    return [0x06, 0x01, 0x29, 0x00, 0x00, 0x2e]
                elif payload == [0x06, 0x01, 0x2f, 0xff, 0xff, 0x28]:
                    return [0x06, 0x01, 0x2f, 0x0b, 0xd6, 0xf5]
                elif payload == [0x06, 0x01, 0x30, 0xff, 0xff, 0x37]:
                    return [0x06, 0x01, 0x30, 0x03, 0x20, 0x14]
                elif payload == [0x06, 0x01, 0x00, 0xff, 0xff, 0x07]:
                    return [0x06, 0x01, 0x00, 0x00, 0x00, 0x07]
                else:
                    raise NotImplementedError
            else:
                raise NotImplementedError

    def register(self) -> SAInfoRegisterPacket:
        """Send a Register request packet for SA."""
        payload = SAInfoRequestPacket.create(
            sa_info_type=SARegisterServiceIDEnum.REGISTRATION
        ).to_pdu()
        response = self.request(payload=payload)
        device = SAInfoRegisterPacket.from_pdu(pdu=response)
        self._device = device
        return device

    def read_state(self, service_id) -> SAStateReadResponsePacket:
        """Send a State Read request packet for SA."""
        payload = SAStateReadRequestPacket.create(
            type_id=self._device.type_id,
            service_id=service_id
        ).to_pdu()
        response = self.request(payload=payload)
        print(f'response: {response}')
        response = SAStateReadResponsePacket.from_pdu(pdu=response)
        assert isinstance(response, SAStateReadResponsePacket)
        return response

    # def read_sensor(self, service_id) -> SASensorReadMultiResponsePacket:
    #     """Send a Sensor Read request for SA."""
    #     raise NotImplementedError

    def write_state(self, service_id, value) -> SAStateWriteResponsePacket:
        """Send a State Write request packet for SA."""
        payload = SAStateWriteRequestPacket.create(
            type_id=self._device.type_id,
            service_id=service_id,
            value=value
        ).to_pdu()
        response = self.request(payload=payload)
        response = SAStateWriteResponsePacket.from_pdu(pdu=response)
        assert isinstance(response, SAStateWriteResponsePacket)
        return response


__all__ = ['SmartApplication']



import setuptools
import re

about = {}
with open('twseia/__version__.py') as fh:
    for line in fh.readlines():
        about.update(dict(re.findall('__([a-z]+)__ *= *\'([^"]+)\'', line)))

setuptools.setup(
    name=about.get('title'),
    version=about.get('version'),
    description=about.get('description'),
    author=about.get('author'),
    author_email=about.get('author_email'),
    url=about.get('url'),
    license=about.get('license'),
    packages=['twseia'],
    install_requires=['pyserial'],
    classifier=[
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy'
    ]
)



import enum
import logging
from .services import UInt8Service
from .services import Int8Service
from .services import UInt16Service
from .services import HMService
from .services import Enum16BitService
from .services import MDService
from .services import SAServiceBase
from .services import Enum16Service

logger = logging.getLogger(__name__)


class SADevice:

    @classmethod
    def convert_dev_specific_service(cls, pdu: list, is_fixed_len_pdu: bool) -> SAServiceBase:
        raise NotImplementedError


class ACPowerService(enum.IntEnum):
    OFF = 0
    ON = 1


class ACOpModeService(enum.IntEnum):
    COOL = 0
    DEHUMIDIFICATION = 1
    FAN = 2
    AUTO = 3
    HEAT = 4


class ACFanLevelService(enum.IntEnum):
    AUTO = 0
    LEVEL_1 = 1
    LEVEL_2 = 2
    LEVEL_3 = 3
    LEVEL_4 = 4
    LEVEL_5 = 5
    LEVEL_6 = 6
    LEVEL_7 = 7
    LEVEL_8 = 8
    LEVEL_9 = 9
    LEVEL_10 = 10
    LEVEL_11 = 11
    LEVEL_12 = 12
    LEVEL_13 = 13
    LEVEL_14 = 14
    LEVEL_15 = 15


# class ACTemperatureCfgService(enum.IntEnum):
# class ACTemperatureService(enum.IntEnum):
class ACComfortableService(enum.IntEnum):
    OFF = 0
    ON = 1


# class ACComfortableTimerService(enum.IntEnum):
class ACFuzzyTemperatureService(enum.IntEnum):
    BALANCE = 0
    COLD = 1
    HOT = 2
    OFF = 3
    ON = 4


class ACAirCleanModeService(enum.IntEnum):
    OFF = 0
    ON = 1


# class ACClockOnService(enum.IntEnum):
# class ACClockOffService(enum.IntEnum):
# class ACTimerOnService(enum.IntEnum):
# class ACTimerOffService(enum.IntEnum):
# class ACSysTimeService(enum.IntEnum):
class ACFanUpdownService(enum.IntEnum):
    OFF = 0
    ON = 1


class ACFanUpdownLevelService(enum.IntEnum):
    LEVEL_0 = 0
    LEVEL_1 = 1
    LEVEL_2 = 2
    LEVEL_3 = 3
    LEVEL_4 = 4
    LEVEL_5 = 5
    LEVEL_6 = 6
    LEVEL_7 = 7
    LEVEL_8 = 8
    LEVEL_9 = 9
    LEVEL_10 = 10
    LEVEL_11 = 11
    LEVEL_12 = 12
    LEVEL_13 = 13
    LEVEL_14 = 14
    LEVEL_15 = 15


class ACFanSwingService(enum.IntEnum):
    OFF = 0
    ON = 1


class ACFanSwingLevelService(enum.IntEnum):
    LEVEL_0 = 0
    LEVEL_1 = 1
    LEVEL_2 = 2
    LEVEL_3 = 3
    LEVEL_4 = 4
    LEVEL_5 = 5
    LEVEL_6 = 6
    LEVEL_7 = 7
    LEVEL_8 = 8
    LEVEL_9 = 9
    LEVEL_10 = 10
    LEVEL_11 = 11
    LEVEL_12 = 12
    LEVEL_13 = 13
    LEVEL_14 = 14
    LEVEL_15 = 15


class ACFilterCleanNotifyService(enum.IntEnum):
    NORMAL = 0
    NEED_CLEAN = 1


# class ACDehumidifierCfgService(enum.IntEnum):
# class ACHumidityService(enum.IntEnum):
class ACSysCheckService(enum.IntEnum):
    NORMAL = 0
    ERR_1 = 1
    ERR_2 = 2
    ERR_3 = 3
    ERR_4 = 4
    ERR_5 = 5
    ERR_6 = 6
    ERR_7 = 7
    ERR_8 = 8
    ERR_9 = 9
    ERR_10 = 10
    ERR_11 = 11
    ERR_12 = 12
    ERR_13 = 13
    ERR_14 = 14
    ERR_15 = 15


class ACAirDetectService(enum.IntEnum):
    OFF = 0
    ON = 1


class ACDevMildewService(enum.IntEnum):
    OFF = 0
    ON = 1


class ACSelfCleanService(enum.IntEnum):
    OFF = 0
    ON = 1


class ACMotionDetectModeService(enum.IntEnum):
    OFF = 0
    DETECT_PERSON = 1
    NOT_DETECT_PERSON = 2
    AUTO_DETECTION = 3


class ACFastOpService(enum.IntEnum):
    OFF = 0
    ON = 1


class ACPowerSavingOpService(enum.IntEnum):
    OFF = 0
    ON = 1


# class ACPowerLimitOpService(enum.IntEnum):
class ACRemoteCtrlLockService(enum.IntEnum):
    POWER = 0
    OP_MODE = 1
    TEMPERATURE_CFG = 2
    FAN_UP_DOWN = 3
    FAN_UPDOWN_SWING = 4
    OFF_KEY = 5


class ACSaaCtrlAudioService(enum.IntEnum):
    AUDIO = 0
    SILENT = 1


class ACBodyDisplayModeService(enum.IntEnum):
    BRIGHT = 0
    DARK = 1
    OFF = 2
    ALL_OFF = 3


class ACMoisturizeModeService(enum.IntEnum):
    OFF = 0
    LOW = 1
    HIGH = 2


# class ACOutdoorTemperatureService(enum.IntEnum):
# class ACIndoorUnitWattService(enum.IntEnum):
# class ACOutdoorUnitWattService(enum.IntEnum):
# class ACOutdoorUnitCurrentService(enum.IntEnum):
# class ACOutdoorUnitVoltateService(enum.IntEnum):
# class ACOutdoorUnitPowerFactorService(enum.IntEnum):
# class ACOutdoorUnitInstantWattService(enum.IntEnum):
# class ACTotalWattService(enum.IntEnum):
class ACDisplayErrService(enum.IntEnum):
    NORMAL = 0
    ERR_1 = 1
    ERR_2 = 2
    ERR_3 = 3
    ERR_4 = 4
    ERR_5 = 5
    ERR_6 = 6
    ERR_7 = 7
    ERR_8 = 8
    ERR_9 = 9
    ERR_10 = 10
    ERR_11 = 11
    ERR_12 = 12
    ERR_13 = 13
    ERR_14 = 14
    ERR_15 = 15


class ACErrHistory1Service(enum.IntEnum):
    NORMAL = 0
    ERR_1 = 1
    ERR_2 = 2
    ERR_3 = 3
    ERR_4 = 4
    ERR_5 = 5
    ERR_6 = 6
    ERR_7 = 7
    ERR_8 = 8
    ERR_9 = 9
    ERR_10 = 10
    ERR_11 = 11
    ERR_12 = 12
    ERR_13 = 13
    ERR_14 = 14
    ERR_15 = 15


class ACErrHistory2Service(enum.IntEnum):
    NORMAL = 0
    ERR_1 = 1
    ERR_2 = 2
    ERR_3 = 3
    ERR_4 = 4
    ERR_5 = 5
    ERR_6 = 6
    ERR_7 = 7
    ERR_8 = 8
    ERR_9 = 9
    ERR_10 = 10
    ERR_11 = 11
    ERR_12 = 12
    ERR_13 = 13
    ERR_14 = 14
    ERR_15 = 15


class ACErrHistory3Service(enum.IntEnum):
    NORMAL = 0
    ERR_1 = 1
    ERR_2 = 2
    ERR_3 = 3
    ERR_4 = 4
    ERR_5 = 5
    ERR_6 = 6
    ERR_7 = 7
    ERR_8 = 8
    ERR_9 = 9
    ERR_10 = 10
    ERR_11 = 11
    ERR_12 = 12
    ERR_13 = 13
    ERR_14 = 14
    ERR_15 = 15


class ACErrHistory4Service(enum.IntEnum):
    NORMAL = 0
    ERR_1 = 1
    ERR_2 = 2
    ERR_3 = 3
    ERR_4 = 4
    ERR_5 = 5
    ERR_6 = 6
    ERR_7 = 7
    ERR_8 = 8
    ERR_9 = 9
    ERR_10 = 10
    ERR_11 = 11
    ERR_12 = 12
    ERR_13 = 13
    ERR_14 = 14
    ERR_15 = 15


class ACErrHistory5Service(enum.IntEnum):
    NORMAL = 0
    ERR_1 = 1
    ERR_2 = 2
    ERR_3 = 3
    ERR_4 = 4
    ERR_5 = 5
    ERR_6 = 6
    ERR_7 = 7
    ERR_8 = 8
    ERR_9 = 9
    ERR_10 = 10
    ERR_11 = 11
    ERR_12 = 12
    ERR_13 = 13
    ERR_14 = 14
    ERR_15 = 15


# class ACMaintenanceAccuOpHourService(enum.IntEnum):
# class ACFilterAccuOpHourService(enum.IntEnum):
# class ACSysYearService(enum.IntEnum):
# class ACSysMonthDayService(enum.IntEnum):
# class ACMonthlyWattService(enum.IntEnum):
# class ACTimerOff2Service(enum.IntEnum):


class ACServiceIDEnum(enum.IntEnum):
    POWER_RW = 0x00
    OP_MODE_RW = 0x01
    FAN_LEVEL_RW = 0x02
    TEMPERATURE_CFG_RW = 0x03
    TEMPERATURE_R = 0x04
    COMFORTABLE_RW = 0x05
    COMFORTABLE_TIMER_RW = 0x06
    FUZZY_TEMPERATURE_RW = 0x07
    AIR_CLEAN_MODE_RW = 0x08
    CLOCK_ON_RW = 0x09
    CLOCK_OFF_RW = 0x0a
    TIMER_ON_RW = 0x0b
    TIMER_OFF_RW = 0x0c
    SYS_TIME_RW = 0x0d
    FAN_UPDOWN_RW = 0x0e
    FAN_UPDOWN_LEVEL_RW = 0x0f
    FAN_SWING_RW = 0x10
    FAN_SWING_LEVEL_RW = 0x11
    FILTER_CLEAN_NOTIFY_RW = 0x12
    DEHUMIDIFIER_CFG_RW = 0x13
    HUMIDITY_R = 0x14
    SYS_CHECK_R = 0x15
    AIR_DETECT_RW = 0x16
    DEV_MILDEW_RW = 0x17
    SELF_CLEAN_RW = 0x18
    MOTION_DETECT_MODE_RW = 0x19
    FAST_OP_RW = 0x1a
    POWER_SAVING_OP_RW = 0x1b
    POWER_LIMIT_OP_RW = 0x1c
    REMOTE_CTRL_LOCK_RW = 0x1d
    SAA_CTRL_AUDIO_RW = 0x1e
    BODY_DISPLAY_MODE_RW = 0x1f
    MOISTURIZE_MODE_RW = 0x20
    OUTDOOR_TEMPERATURE_R = 0x21
    INDOOR_UNIT_WATT_R = 0x22
    OUTDOOR_UNIT_WATT_R = 0x23
    OUTDOOR_UNIT_CURRENT_R = 0x24
    OUTDOOR_UNIT_VOLTATE_R = 0x25
    OUTDOOR_UNIT_POWER_FACTOR_R = 0x26
    OUTDOOR_UNIT_INSTANT_WATT_R = 0x27
    TOTAL_WATT_RW = 0x28
    DISPLAY_ERR_R = 0x29
    ERR_HISTORY_1_R = 0x2a
    ERR_HISTORY_2_R = 0x2b
    ERR_HISTORY_3_R = 0x2c
    ERR_HISTORY_4_R = 0x2d
    ERR_HISTORY_5_R = 0x2e
    MAINTENANCE_ACCU_OP_HOUR_RW = 0x2f
    FILTER_ACCU_OP_HOUR_RW = 0x30
    SYS_YEAR_RW = 0x31
    SYS_MONTH_DAY_RW = 0x32
    MONTHLY_WATT_R = 0x33
    TIMER_OFF_2_RW = 0x34


class AirConditioner(SADevice):
    @classmethod
    def convert_dev_specific_service(cls, pdu: list, is_fixed_len_pdu: bool) -> SAServiceBase:
        if not isinstance(pdu, list) or len(pdu) < 3:
            raise ValueError(f'service pdu invalid, {pdu}')
        if not is_fixed_len_pdu:
            raise NotImplementedError

        _pdu = list(pdu)
        logger.debug(f'parsing dev service pdu {_pdu}')
        _service = SAServiceBase.from_fixed_len_pdu(pdu=_pdu)
        if _service.service_id == ACServiceIDEnum.POWER_RW:
            _service = Enum16Service.from_fixed_len_pdu(pdu=_pdu)
        elif _service.service_id == ACServiceIDEnum.OP_MODE_RW:
            _service = Enum16Service.from_fixed_len_pdu(pdu=_pdu)
        elif _service.service_id == ACServiceIDEnum.FAN_LEVEL_RW:
            _service = Enum16Service.from_fixed_len_pdu(pdu=_pdu)
        elif _service.service_id == ACServiceIDEnum.TEMPERATURE_CFG_RW:
            _service = UInt8Service.from_fixed_len_pdu(pdu=_pdu)
        elif _service.service_id == ACServiceIDEnum.TEMPERATURE_R:
            _service = Int8Service.from_fixed_len_pdu(pdu=_pdu)
        elif _service.service_id == ACServiceIDEnum.COMFORTABLE_RW:
            _service = Enum16Service.from_fixed_len_pdu(pdu=_pdu)
        elif _service.service_id == ACServiceIDEnum.COMFORTABLE_TIMER_RW:
            _service = UInt16Service.from_fixed_len_pdu(pdu=_pdu)
        elif _service.service_id == ACServiceIDEnum.FUZZY_TEMPERATURE_RW:
            _service = Enum16Service.from_fixed_len_pdu(pdu=_pdu)
        elif _service.service_id == ACServiceIDEnum.AIR_CLEAN_MODE_RW:
            _service = Enum16Service.from_fixed_len_pdu(pdu=_pdu)
        elif _service.service_id == ACServiceIDEnum.CLOCK_ON_RW:
            _service = HMService.from_fixed_len_pdu(pdu=_pdu)
        elif _service.service_id == ACServiceIDEnum.CLOCK_OFF_RW:
            _service = HMService.from_fixed_len_pdu(pdu=_pdu)
        elif _service.service_id == ACServiceIDEnum.TIMER_ON_RW:
            _service = UInt16Service.from_fixed_len_pdu(pdu=_pdu)
        elif _service.service_id == ACServiceIDEnum.TIMER_OFF_RW:
            _service = UInt16Service.from_fixed_len_pdu(pdu=_pdu)
        elif _service.service_id == ACServiceIDEnum.SYS_TIME_RW:
            _service = HMService.from_fixed_len_pdu(pdu=_pdu)
        elif _service.service_id == ACServiceIDEnum.FAN_UPDOWN_RW:
            _service = Enum16Service.from_fixed_len_pdu(pdu=_pdu)
        elif _service.service_id == ACServiceIDEnum.FAN_UPDOWN_LEVEL_RW:
            _service = Enum16Service.from_fixed_len_pdu(pdu=_pdu)
        elif _service.service_id == ACServiceIDEnum.FAN_SWING_RW:
            _service = Enum16Service.from_fixed_len_pdu(pdu=_pdu)
        elif _service.service_id == ACServiceIDEnum.FAN_SWING_LEVEL_RW:
            _service = Enum16Service.from_fixed_len_pdu(pdu=_pdu)
        elif _service.service_id == ACServiceIDEnum.FILTER_CLEAN_NOTIFY_RW:
            _service = Enum16Service.from_fixed_len_pdu(pdu=_pdu)
        elif _service.service_id == ACServiceIDEnum.DEHUMIDIFIER_CFG_RW:
            _service = UInt8Service.from_fixed_len_pdu(pdu=_pdu)
        elif _service.service_id == ACServiceIDEnum.HUMIDITY_R:
            _service = UInt8Service.from_fixed_len_pdu(pdu=_pdu)
        elif _service.service_id == ACServiceIDEnum.SYS_CHECK_R:
            _service = Enum16BitService.from_fixed_len_pdu(pdu=_pdu)
        elif _service.service_id == ACServiceIDEnum.AIR_DETECT_RW:
            _service = Enum16Service.from_fixed_len_pdu(pdu=_pdu)
        elif _service.service_id == ACServiceIDEnum.DEV_MILDEW_RW:
            _service = Enum16Service.from_fixed_len_pdu(pdu=_pdu)
        elif _service.service_id == ACServiceIDEnum.SELF_CLEAN_RW:
            _service = Enum16Service.from_fixed_len_pdu(pdu=_pdu)
        elif _service.service_id == ACServiceIDEnum.MOTION_DETECT_MODE_RW:
            _service = Enum16Service.from_fixed_len_pdu(pdu=_pdu)
        elif _service.service_id == ACServiceIDEnum.FAST_OP_RW:
            _service = Enum16Service.from_fixed_len_pdu(pdu=_pdu)
        elif _service.service_id == ACServiceIDEnum.POWER_SAVING_OP_RW:
            _service = Enum16Service.from_fixed_len_pdu(pdu=_pdu)
        elif _service.service_id == ACServiceIDEnum.POWER_LIMIT_OP_RW:
            _service = UInt16Service.from_fixed_len_pdu(pdu=_pdu)
        elif _service.service_id == ACServiceIDEnum.REMOTE_CTRL_LOCK_RW:
            _service = Enum16BitService.from_fixed_len_pdu(pdu=_pdu)
        elif _service.service_id == ACServiceIDEnum.SAA_CTRL_AUDIO_RW:
            _service = Enum16Service.from_fixed_len_pdu(pdu=_pdu)
        elif _service.service_id == ACServiceIDEnum.BODY_DISPLAY_MODE_RW:
            _service = Enum16Service.from_fixed_len_pdu(pdu=_pdu)
        elif _service.service_id == ACServiceIDEnum.MOISTURIZE_MODE_RW:
            _service = Enum16Service.from_fixed_len_pdu(pdu=_pdu)
        elif _service.service_id == ACServiceIDEnum.OUTDOOR_TEMPERATURE_R:
            _service = Int8Service.from_fixed_len_pdu(pdu=_pdu)
        elif _service.service_id == ACServiceIDEnum.INDOOR_UNIT_WATT_R:
            _service = UInt16Service.from_fixed_len_pdu(pdu=_pdu)
        elif _service.service_id == ACServiceIDEnum.OUTDOOR_UNIT_WATT_R:
            _service = UInt16Service.from_fixed_len_pdu(pdu=_pdu)
        elif _service.service_id == ACServiceIDEnum.OUTDOOR_UNIT_CURRENT_R:
            _service = UInt16Service.from_fixed_len_pdu(pdu=_pdu)
        elif _service.service_id == ACServiceIDEnum.OUTDOOR_UNIT_VOLTATE_R:
            _service = UInt16Service.from_fixed_len_pdu(pdu=_pdu)
        elif _service.service_id == ACServiceIDEnum.OUTDOOR_UNIT_POWER_FACTOR_R:
            _service = UInt16Service.from_fixed_len_pdu(pdu=_pdu)
        elif _service.service_id == ACServiceIDEnum.OUTDOOR_UNIT_INSTANT_WATT_R:
            _service = UInt16Service.from_fixed_len_pdu(pdu=_pdu)
        elif _service.service_id == ACServiceIDEnum.TOTAL_WATT_RW:
            _service = UInt16Service.from_fixed_len_pdu(pdu=_pdu)
        elif _service.service_id == ACServiceIDEnum.DISPLAY_ERR_R:
            _service = Enum16BitService.from_fixed_len_pdu(pdu=_pdu)
        elif _service.service_id == ACServiceIDEnum.ERR_HISTORY_1_R:
            _service = Enum16BitService.from_fixed_len_pdu(pdu=_pdu)
        elif _service.service_id == ACServiceIDEnum.ERR_HISTORY_2_R:
            _service = Enum16BitService.from_fixed_len_pdu(pdu=_pdu)
        elif _service.service_id == ACServiceIDEnum.ERR_HISTORY_3_R:
            _service = Enum16BitService.from_fixed_len_pdu(pdu=_pdu)
        elif _service.service_id == ACServiceIDEnum.ERR_HISTORY_4_R:
            _service = Enum16BitService.from_fixed_len_pdu(pdu=_pdu)
        elif _service.service_id == ACServiceIDEnum.ERR_HISTORY_5_R:
            _service = Enum16BitService.from_fixed_len_pdu(pdu=_pdu)
        elif _service.service_id == ACServiceIDEnum.MAINTENANCE_ACCU_OP_HOUR_RW:
            _service = UInt16Service.from_fixed_len_pdu(pdu=_pdu)
        elif _service.service_id == ACServiceIDEnum.FILTER_ACCU_OP_HOUR_RW:
            _service = UInt16Service.from_fixed_len_pdu(pdu=_pdu)
        elif _service.service_id == ACServiceIDEnum.SYS_YEAR_RW:
            _service = UInt16Service.from_fixed_len_pdu(pdu=_pdu)
        elif _service.service_id == ACServiceIDEnum.SYS_MONTH_DAY_RW:
            _service = MDService.from_fixed_len_pdu(pdu=_pdu)
        elif _service.service_id == ACServiceIDEnum.MONTHLY_WATT_R:
            _service = UInt16Service.from_fixed_len_pdu(pdu=_pdu)
        elif _service.service_id == ACServiceIDEnum.TIMER_OFF_2_RW:
            _service = UInt16Service.from_fixed_len_pdu(pdu=_pdu)
        else:
            logger.warning(f'Unknown {cls.__name__} service_id {_service.service_id}')
            _service = Enum16Service.from_fixed_len_pdu(pdu=pdu)

        _service.name = ACServiceIDEnum(_service.service_id).name
        return _service


class DehumidifierServiceIDEnum(enum.IntEnum):
    POWER_RW = 0x00
    OP_MODE_RW = 0x01
    OP_TIMER_RW = 0x02
    HUMIDITY_CFG_RW = 0x03
    DEHUMIDIFIER_LEVEL_RW = 0x04
    DRY_CLOTHE_LEVEL_RW = 0x05
    TEMPERATURE_R = 0x06
    HUMIDITY_R = 0x07
    FAN_DIRECTION_AUTO_RW = 0x08
    FAN_DIRECTION_LEVEL_RW = 0x09
    WATER_FULL_ALARM_R = 0x0a
    FILTER_CLEAN_NOTIFY_RW = 0x0b
    MOOD_LED_RW = 0x0c
    AIR_CLEAN_MODE_RW = 0x0d
    FAN_LEVEL_RW = 0x0e
    SIDE_FAN_R = 0x0f
    AUDIO_RW = 0x10
    DEFROST_DISPLAY_R = 0x11
    DISPLAY_ERR_R = 0x12
    DEV_MILDEW_RW = 0x13
    HUMIDITY_HIGH_NOTIFY_RW = 0x14
    HUMIDITY_HIGH_CFG_RW = 0x15
    KEYPAD_LOCK_RW = 0x16
    REMOTE_CTRL_LOCK_RW = 0x17
    SAA_CTRL_AUDIO_RW = 0x18
    OP_CURRENT_R = 0x19
    OP_VOLTAGE_R = 0x1a
    OP_POWER_FACTOR_R = 0x1b
    OP_POWER_WATT_RW = 0x1c
    TOTAL_WATT_RW = 0x1d
    ERR_HISTORY_1_R = 0x1e
    ERR_HISTORY_2_R = 0x1f
    ERR_HISTORY_3_R = 0x20
    ERR_HISTORY_4_R = 0x21
    ERR_HISTORY_5_R = 0x22


class Dehumidifier(SADevice):
    @classmethod
    def convert_dev_specific_service(cls, pdu: list, is_fixed_len_pdu: bool) -> SAServiceBase:
        if not isinstance(pdu, list) or len(pdu) < 3:
            raise ValueError(f'service pdu invalid, {pdu}')
        if not is_fixed_len_pdu:
            raise NotImplementedError

        _pdu = list(pdu)
        _service = SAServiceBase.from_fixed_len_pdu(pdu=_pdu)
        if _service.service_id == DehumidifierServiceIDEnum.POWER_RW:
            _service = Enum16Service.from_fixed_len_pdu(pdu=_pdu)
        elif _service.service_id == DehumidifierServiceIDEnum.OP_MODE_RW:
            _service = Enum16Service.from_fixed_len_pdu(pdu=_pdu)
        elif _service.service_id == DehumidifierServiceIDEnum.OP_TIMER_RW:
            _service = UInt8Service.from_fixed_len_pdu(pdu=_pdu)
        elif _service.service_id == DehumidifierServiceIDEnum.HUMIDITY_CFG_RW:
            _service = UInt8Service.from_fixed_len_pdu(pdu=_pdu)
        elif _service.service_id == DehumidifierServiceIDEnum.DEHUMIDIFIER_LEVEL_RW:
            _service = Enum16Service.from_fixed_len_pdu(pdu=_pdu)
        elif _service.service_id == DehumidifierServiceIDEnum.DRY_CLOTHE_LEVEL_RW:
            _service = Enum16Service.from_fixed_len_pdu(pdu=_pdu)
        elif _service.service_id == DehumidifierServiceIDEnum.TEMPERATURE_R:
            _service = Int8Service.from_fixed_len_pdu(pdu=_pdu)
        elif _service.service_id == DehumidifierServiceIDEnum.HUMIDITY_R:
            _service = UInt8Service.from_fixed_len_pdu(pdu=_pdu)
        elif _service.service_id == DehumidifierServiceIDEnum.FAN_DIRECTION_AUTO_RW:
            _service = Enum16Service.from_fixed_len_pdu(pdu=_pdu)
        elif _service.service_id == DehumidifierServiceIDEnum.FAN_DIRECTION_LEVEL_RW:
            _service = Enum16Service.from_fixed_len_pdu(pdu=_pdu)
        elif _service.service_id == DehumidifierServiceIDEnum.WATER_FULL_ALARM_R:
            _service = Enum16Service.from_fixed_len_pdu(pdu=_pdu)
        elif _service.service_id == DehumidifierServiceIDEnum.FILTER_CLEAN_NOTIFY_RW:
            _service = Enum16Service.from_fixed_len_pdu(pdu=_pdu)
        elif _service.service_id == DehumidifierServiceIDEnum.MOOD_LED_RW:
            _service = Enum16Service.from_fixed_len_pdu(pdu=_pdu)
        elif _service.service_id == DehumidifierServiceIDEnum.AIR_CLEAN_MODE_RW:
            _service = Enum16Service.from_fixed_len_pdu(pdu=_pdu)
        elif _service.service_id == DehumidifierServiceIDEnum.FAN_LEVEL_RW:
            _service = Enum16Service.from_fixed_len_pdu(pdu=_pdu)
        elif _service.service_id == DehumidifierServiceIDEnum.SIDE_FAN_R:
            _service = Enum16Service.from_fixed_len_pdu(pdu=_pdu)
        elif _service.service_id == DehumidifierServiceIDEnum.AUDIO_RW:
            _service = Enum16Service.from_fixed_len_pdu(pdu=_pdu)
        elif _service.service_id == DehumidifierServiceIDEnum.DEFROST_DISPLAY_R:
            _service = Enum16Service.from_fixed_len_pdu(pdu=_pdu)
        elif _service.service_id == DehumidifierServiceIDEnum.DISPLAY_ERR_R:
            _service = Enum16BitService.from_fixed_len_pdu(pdu=_pdu)
        elif _service.service_id == DehumidifierServiceIDEnum.DEV_MILDEW_RW:
            _service = Enum16Service.from_fixed_len_pdu(pdu=_pdu)
        elif _service.service_id == DehumidifierServiceIDEnum.HUMIDITY_HIGH_NOTIFY_RW:
            _service = Enum16Service.from_fixed_len_pdu(pdu=_pdu)
        elif _service.service_id == DehumidifierServiceIDEnum.HUMIDITY_HIGH_CFG_RW:
            _service = UInt8Service.from_fixed_len_pdu(pdu=_pdu)
        elif _service.service_id == DehumidifierServiceIDEnum.KEYPAD_LOCK_RW:
            _service = Enum16Service.from_fixed_len_pdu(pdu=_pdu)
        elif _service.service_id == DehumidifierServiceIDEnum.REMOTE_CTRL_LOCK_RW:
            _service = Enum16BitService.from_fixed_len_pdu(pdu=_pdu)
        elif _service.service_id == DehumidifierServiceIDEnum.SAA_CTRL_AUDIO_RW:
            _service = Enum16Service.from_fixed_len_pdu(pdu=_pdu)
        elif _service.service_id == DehumidifierServiceIDEnum.OP_CURRENT_R:
            _service = UInt16Service.from_fixed_len_pdu(pdu=_pdu)
        elif _service.service_id == DehumidifierServiceIDEnum.OP_VOLTAGE_R:
            _service = UInt16Service.from_fixed_len_pdu(pdu=_pdu)
        elif _service.service_id == DehumidifierServiceIDEnum.OP_POWER_FACTOR_R:
            _service = UInt16Service.from_fixed_len_pdu(pdu=_pdu)
        elif _service.service_id == DehumidifierServiceIDEnum.OP_POWER_WATT_RW:
            _service = UInt16Service.from_fixed_len_pdu(pdu=_pdu)
        elif _service.service_id == DehumidifierServiceIDEnum.TOTAL_WATT_RW:
            _service = UInt16Service.from_fixed_len_pdu(pdu=_pdu)
        elif _service.service_id == DehumidifierServiceIDEnum.ERR_HISTORY_1_R:
            _service = Enum16BitService.from_fixed_len_pdu(pdu=_pdu)
        elif _service.service_id == DehumidifierServiceIDEnum.ERR_HISTORY_2_R:
            _service = Enum16BitService.from_fixed_len_pdu(pdu=_pdu)
        elif _service.service_id == DehumidifierServiceIDEnum.ERR_HISTORY_3_R:
            _service = Enum16BitService.from_fixed_len_pdu(pdu=_pdu)
        elif _service.service_id == DehumidifierServiceIDEnum.ERR_HISTORY_4_R:
            _service = Enum16BitService.from_fixed_len_pdu(pdu=_pdu)
        elif _service.service_id == DehumidifierServiceIDEnum.ERR_HISTORY_5_R:
            _service = Enum16BitService.from_fixed_len_pdu(pdu=_pdu)
        else:
            raise NotImplementedError

        _service.name = ACServiceIDEnum(_service.service_id).name
        return _service


__all__ = [
    'AirConditioner',
    'ACServiceIDEnum',
    'Dehumidifier',
    'DehumidifierServiceIDEnum'
]
